from __future__ import print_function, unicode_literals
import sys, os

from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.containers import VSplit, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import print_container
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.styles import Style

instructions = """
    1. To trigger auto-completion, start with TAB or type the beginning of a command or option and press Tab.
    2. Use the arrow keys or Tab to navigate through the suggestions.
    3. Press Enter to accept a suggestion or Esc to cancel.
    4. If an option requires a value, use --option=value instead of --option value.
    5. Special key '%%' will change scopes. Use it with TAB completion to change depth levels.
        %% ..        go back a depth level
        %% {scope}   change context to selected scope in TAB completion
        %%           unscope to root of the tree
"""

from prompt_toolkit.layout.containers import (
    HorizontalAlign,
    HSplit,
    VerticalAlign,
    VSplit,
    Window,
    WindowAlign,
)

from prompt_toolkit.widgets import (
    Frame,
    Label,
    TextArea,
)

buffer1 = Buffer()

def is_non_zero_file(fpath):
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0

tip = 'TIP: resource completion coming soon!'
version = '1.0.0'
text = f"Purpose: Cloud Wrapper\n{tip}"
title = HTML("""<b>GOAT INTERFACE</b>""")
func_title = HTML("""<b>Hotkeys</b>""")
misc_title = HTML("""<b>Commands</b>""")
advanced_title = HTML("""<b>Advanced</b>""")
input = """\
[F7]  Xtract CLI Tree
[F8]  Toggle Provider
[F9]  Toggle Profile
[F10] Toggle VIM
[F12] Toggle Safety Mode
"""

misc_input = """\
e|exit    : exit shell
c|clear   : clear screen
h|help    : display usage
##        : key to run OS command
%%        : key to change scopes
"""

def getLayout(with_help=None):
    os.system('clear')

    instructions_text = instructions if with_help else "INFO: general instructions will be displayed here with the 'help' command"
    instructions_area = TextArea(
        text=instructions_text,
        read_only=True,
        scrollbar=with_help
    )

    instructions_frame = Frame(
        instructions_area,
        title=HTML("<b>Instructions</b>")
    )

    # Explicitly setting the width of frames in VSplit
    func_frame = Frame(body=Label(text=input.strip()), width=D(preferred=60), title=(func_title))
    misc_frame = Frame(body=Label(text=misc_input.rstrip()), width=D(preferred=60), title=(misc_title))

    # Aligning frames within VSplit
    main_vsplit = VSplit([
        func_frame,
        misc_frame
    ], align=VerticalAlign.CENTER)

    main_layout = HSplit([
        Frame(
            Window(FormattedTextControl(text), height=D.exact(2), align=WindowAlign.CENTER), title=(title)
        ),
        main_vsplit,
        instructions_frame
    ])

    print_container(main_layout)

if __name__ == '__main__':
    getLayout()



"""
Instead of using the 'az.cli' python package, which download complete Azure-CLI, assume 'az' is available on PATH.
"""
import json
import platform
import re
import subprocess  # nosec B404
from collections import namedtuple
from typing import List

AzCliResult = namedtuple("AzCliResult", ["exit_code", "result_dict", "logs"])


def az(command: str) -> AzCliResult:
    """
    Alternative to 'az.cli' to interact with Azure CLI. This implementation assumes 'az' is available on the Path.
    """
    cmd_list = _prepare_cmd_list(command)
    if platform.system() == "Windows":
        result = subprocess.run(cmd_list, capture_output=True, shell=True)  # noqa: S602 #nosec B602
        encoding = "ISO-8859-1"
    elif platform.system() in ("Linux", "Darwin"):
        result = subprocess.run(cmd_list, capture_output=True)  # noqa: S603 #nosec B603
        encoding = "utf-8"
    else:
        raise RuntimeError(f"Unsupported platform: {platform.system()}")

    return_code = result.returncode

    if return_code != 0:
        return AzCliResult(return_code, None, bytes.decode(result.stderr))

    # PATCH operation does not have result dict.
    if not result.stdout:
        return AzCliResult(return_code, None, None)

    data = bytes.decode(result.stdout, encoding=encoding)

    try:
        return AzCliResult(return_code, json.loads(data), None)
    except json.decoder.JSONDecodeError:
        # Not json? Not oke!
        return AzCliResult(1, None, data)


def _prepare_cmd_list(command: str) -> List[str]:
    """Flatten single-quoted literals into single line and split into list."""
    if not command.startswith("az "):
        command = "az " + command

    literals = re.findall(r"'((?:.|\n)+?)'", command)
    if not literals:
        return command.split()

    for _i, literal in enumerate(literals):
        command = command.replace(f"'{literal}'", f"${_i}", 1)

    parts = command.split()

    for _i, literal in enumerate(literals):
        for _j, part in enumerate(parts):
            if part == f"${_i}":
                parts[_j] = " ".join(literal.split())

    return parts



from prompt_toolkit.completion import Completer, Completion
from fuzzyfinder import fuzzyfinder
import shlex
import json
import os
import asyncio
from pathlib import Path
from . import misc
from toolbox.misc import get_save_path, attempt_load_custom_json

from goatshell.parser import Parser  # Replace this with your actual Parser import

import logging

logger = logging.getLogger(__name__)

class GoatCompleter(Completer):
    def __init__(self, parser):
        self.inline_help = True
        self.min_input_len = 0
        self.parser = parser
        self.current_json = None  # NEW: attribute to store the current JSON
        self.goat_dict = None
        self.current_cloud_provider = None
        self.current_context = []

        # Dictionary mapping command tokens to their descriptions
        self.command_descriptions = {
            "aliyun": "Alibaba Cloud Command Line Interface",
            "aws": "Amazon Web Services",
            "az": "Microsoft Azure Cloud Platform",
            "gcloud": "Google Cloud Platform",
            "goat": "Local Goat Application",
            "oci": "Oracle Cloud Infrastructure",
            "ibmcloud": "IBM Cloud Infrastructure",
            "ovhai": "OVH AI & Public Cloud"
            # Add more commands and descriptions as needed
        }

    def set_current_context(self, context):
        self.current_context = context

    def set_current_cloud_provider(self, provider):
        self.current_cloud_provider = provider

    def load_json(self, provider):
        # Get the JSON path using the modularized function
        user_json_path = Path(get_save_path(provider)).with_suffix('.json')
    
        # If user's custom JSON does not exist, use the default one
        if not user_json_path.exists():
            if not attempt_load_custom_json(provider):
                # If failed to load user's custom JSON, use the system default
                user_json_path = Path(os.path.dirname(os.path.realpath(__file__))) / "data" / f"{provider}.json"

        try:
            with user_json_path.open(encoding='utf-8') as json_file:
                self.goat_dict = json.load(json_file)
    
            self.parser = Parser(str(user_json_path), provider)
        except Exception as ex:
            logger.error(f"Exception while loading JSON for {provider}: {ex}")
    
    def set_current_cloud_provider(self, provider):
        if provider in ["aliyun", "aws", "az", "gcloud", "goat", "oci", "ibmcloud", "ovhai"]:
            self.current_cloud_provider = provider
            self.load_json(provider)
            logger.debug(f"Current cloud provider set to: {provider}")
        else:
            logger.error(f"Invalid cloud provider: {provider}")

    def get_percent_completions(self, trimmed_input, document):
        # Method for generating completions when '%%' is typed
        completions = []
        current_node = self.goat_dict.get(self.current_cloud_provider, {})

        # If there is a current context, navigate to the appropriate node
        for context_part in self.current_context:
            current_node = current_node.get('subcommands', {}).get(context_part, {})

        # Combine all subcommands for fuzzy matching
        subcommand_keys = current_node.get('subcommands', {}).keys()

        # Use fuzzyfinder to filter and rank candidates based on the trimmed input
        fuzzy_matches = fuzzyfinder(trimmed_input, subcommand_keys)

        # Generate completions based on fuzzy matches
        for match in fuzzy_matches:
            display = f'\u2794 {match}'
            # Calculate start_position to overwrite only the part after '%%'
            start_position = -len(trimmed_input) if trimmed_input else 0
            completions.append(Completion(match, start_position=start_position, display=display))

        return completions

    def get_completions(self, document, complete_event, smart_completion=True):
        input_text = document.text_before_cursor.strip()
        tokens = shlex.split(document.text_before_cursor.strip())

        if input_text.startswith('%%'):
            trimmed_input = input_text[2:]  # Remove '%%' from the input
            yield from self.get_percent_completions(trimmed_input, document)
            return
        
        if self.current_context:
            context_completions = self.get_context_specific_completions(tokens, document)
            if context_completions:
                yield from context_completions
            return

        # Check if there's no input and a current cloud provider is set
        if not tokens and self.current_cloud_provider:
            # Load the json for the current cloud provider
            self.load_json(self.current_cloud_provider)

            # Provide completions for the current cloud provider
            subcommands = self.parser.ast.children
            subcommands = sorted(subcommands, key=lambda x: x.node)
            for subcmd in subcommands:
                display = f'\u2794 {subcmd.node}'
                yield Completion(subcmd.node, display=display, display_meta=subcmd.help)
            return

        if tokens and self.current_cloud_provider:
            # Prepend the current cloud provider as the first token
            tokens = [self.current_cloud_provider] + tokens

            # Update the command line based on selected subcommand
            if len(tokens) > 1:
                # Identify the selected subcommand (e.g., the last token)
                selected_subcommand = tokens[-1]

                # Update the command line by replacing the last token with the selected subcommand
                updated_command_line = " ".join(tokens[:-1]) + f" {selected_subcommand}"

        if not tokens:
            # Check if a command is available before suggesting it
            available_commands = ["aliyun", "aws", "az", "gcloud", "goat", "oci", "ibmcloud", "ovhai"]
            for command in available_commands:
                display = f'\u2794 {command}'
                if misc.is_command_available(command):
                    yield Completion(command, display=display, display_meta=self.command_descriptions.get(command, ""))
            return

        first_token = tokens[0]

        if first_token in ["oci", "aws", "gcloud", "az", "goat", "aliyun", "ibmcloud", "ovhai"] and self.current_json != first_token:
            self.load_json(first_token)
            self.current_json = first_token

        if len(tokens) == 1 and first_token in ["oci", "aws", "gcloud", "az", "goat", "aliyun", "ibmcloud", "ovhai"]:
            subcommands = self.parser.ast.children
            # Sort the subcommands alphabetically before yielding them
            subcommands = sorted(subcommands, key=lambda x: x.node)
            for subcmd in subcommands:
                display = f'\u2794 {subcmd.node}'
                yield Completion(subcmd.node, display=display, display_meta=subcmd.help)
            return

        parsed, unparsed, suggestions = self.parser.parse_tokens(tokens)
    
        if suggestions is None:
            logger.error("suggestions are None")
            return
    
        if unparsed:
            last_token = unparsed[-1]
    
            # Check if the last token is a global option with a value
            if last_token.startswith("--") and "=" in last_token:
                global_option, option_value = last_token.split("=")
    
                # Handle the global option and value here, e.g., navigate to the root of the tree
                # Your code to handle the global option and value goes here
    
            elif last_token.startswith("--") or last_token.startswith('-'):
                if not document.text_before_cursor.endswith("--") or not document.text_before_cursor.endswith('-'):
                    option_prefix = last_token
                    completions = fuzzyfinder(option_prefix, suggestions.keys())
                    for key in completions:
                        display = f'\u2794 {key}'
                        yield Completion(key, start_position=-len(option_prefix), display=display, display_meta=suggestions.get(key, ""))
            else:
                completions = fuzzyfinder(last_token, suggestions.keys())
                for key in completions:
                    display = f'\u2794 {key}'
                    yield Completion(key, start_position=-len(last_token), display=display, display_meta=suggestions.get(key, ""))
        else:
            for key in suggestions.keys():
                display = f'\u2794 {key}'
                yield Completion(key, display=display, display_meta=suggestions.get(key, ""))
    
    def get_context_specific_completions(self, tokens, document):
        self.load_json(self.current_cloud_provider)

        current_node = self.goat_dict.get(self.current_cloud_provider, {})
        context_string = ' '.join(self.current_context)

        # Track the full path traversed so far
        traversed_path = list(self.current_context)

        # Navigate based on the current context
        for context_part in self.current_context:
            if 'subcommands' in current_node and context_part in current_node['subcommands']:
                current_node = current_node['subcommands'][context_part]
            else:
                return []

        # Filter out tokens that are redundant with the context
        filtered_tokens = [token for token in tokens if token not in self.current_context]

        # Navigate based on the filtered tokens and update the traversed path
        for token in filtered_tokens:
            if 'subcommands' in current_node and token in current_node['subcommands']:
                current_node = current_node['subcommands'][token]
                traversed_path.append(token)  # Add token to the traversed path

        # Generate completions based on the current_node and the traversed path
        completions = []

        # Get the current input for fuzzy matching
        current_input = document.get_word_before_cursor()

        # Combine subcommands and options for fuzzy matching
        completion_candidates = list(current_node.get('subcommands', {}).keys()) + list(current_node.get('options', {}).keys())

        # Use fuzzyfinder to filter and rank candidates based on the current input
        fuzzy_matches = fuzzyfinder(current_input, completion_candidates)

        # Generate completions based on fuzzy matches
        for match in fuzzy_matches:
            value = current_node.get('subcommands', {}).get(match, {}) or current_node.get('options', {}).get(match, {})
            help_text = value.get('help', 'No description available')
            display_text = ' '.join(traversed_path + [match]).strip()  # Use the full traversed path
            completions.append(Completion(match, start_position=-len(current_input),
                                          display=f'\u2794 {display_text}', display_meta=help_text))

        return completions or []

    async def get_completions_async(self, document, complete_event):
        logger.debug("Entering get_completions")
        completions = await asyncio.to_thread(self.get_completions, document, complete_event)
        for completion in completions:
            yield completion



from prompt_toolkit import PromptSession

# Define a custom PromptSession class
class DynamicPromptSession(PromptSession):
    def __init__(self, vi_mode_enabled=True, *args, **kwargs):
        self.vi_mode_enabled = vi_mode_enabled
        super().__init__(*args, **kwargs)

    def prompt(self, *args, **kwargs):
        # Set the 'vi_mode' parameter based on 'vi_mode_enabled'
        kwargs['vi_mode'] = self.vi_mode_enabled
        return super().prompt(*args, **kwargs)



from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style
from goatshell.style import styles

def create_toolbar(profile, prefix, vi_mode_enabled, safety_mode_enabled, last_executed_command=None, status_text="", warning_message=None):
    upper_profile = profile.upper()
    upper_prefix = prefix.upper()
    vi_mode_text = "ON" if vi_mode_enabled else "OFF"
    safety_mode_text = "ON" if safety_mode_enabled else "OFF"


    toolbar_class = 'bottom-toolbar-red' if warning_message else 'bottom-toolbar'

    if warning_message:
        toolbar_parts = [ 
               ('class:' + toolbar_class, f'WARNING: {warning_message}')
        ]
    else:
        toolbar_parts = [
                ('class:' + toolbar_class, f'F8 Cloud: {upper_prefix}   F9 Profile: {upper_profile}   F10 VIM {vi_mode_text}   F12 Safe Mode: {safety_mode_text}')
        ]

    if last_executed_command:
        if status_text == "failure":
            toolbar_parts.append(('class:bottom-toolbar', f' | Last Executed: {status_text} => '))
            toolbar_parts.append(('class:failure-text', f'{last_executed_command}'))
        else:
            toolbar_parts.append(('class:bottom-toolbar', ' | Last Executed: '))
            toolbar_parts.append(('class:success-text', f'{last_executed_command}'))

    return FormattedText(toolbar_parts)




import configparser
import os

class SettingsManager:
    def __init__(self):
        self.shell_dir = os.path.expanduser("~/goat/shell/")
        self.config_path = os.path.join(self.shell_dir, 'config.ini')
        self.config = configparser.ConfigParser()
        self._load_config()

    def _load_config(self):
        # Load existing settings if they exist
        if os.path.exists(self.config_path):
            self.config.read(self.config_path)
        else:
            # Create default settings section if it doesn't exist
            self.config['Settings'] = {}

    def save_setting(self, setting_key, setting_value):
        """Save a setting to config.ini."""
        self.config['Settings'][setting_key] = str(setting_value)
        self._save_config()

    def load_setting(self, setting_key, default_value=None):
        """Load a setting from config.ini."""
        return self.config['Settings'].get(setting_key, default_value)

    def _save_config(self):
        # Create directory if it doesn't exist
        if not os.path.exists(self.shell_dir):
            os.makedirs(self.shell_dir)

        # Save the configuration file
        with open(self.config_path, 'w') as configfile:
            self.config.write(configfile)




from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from goatshell.ui import getLayout
import os

class CustomKeyBindings(KeyBindings):
    def __init__(self, app, goatshell_instance):
        super().__init__()
        self.app = app  # Store the Application reference
        self.goatshell_instance = goatshell_instance

        @self.add(Keys.F7)
        def handle_f7(event):
            print()
            cloud_provider = self.goatshell_instance.prefix
            if cloud_provider != 'goat':
                os_command = f"goat {cloud_provider} extract commands"
                try:
                    result = self.goatshell_instance.execute_os_command(os_command)
                    if result == "failure":
                        print("INFO: failed to execute the command.")
                    else:
                        print(f"INFO: executed the command {os_command}")
                except:
                    pass
            else:
                directory = os.path.dirname(os.path.abspath(__file__))
                os_command = f"python3 {directory}/fetch_goat.py"
                try:
                    os.system(os_command)
                except:
                    pass

            event.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to re-prompt.

        @self.add(Keys.F8)
        def handle_f8(event):
            goatshell_instance.switch_to_next_provider()
            if self.goatshell_instance.prefix == 'az':
                sub_id, sub_name = self.goatshell_instance.fetch_current_azure_subscription()
                if sub_id and sub_name:
                    self.goatshell_instance.profile = sub_name
                else:
                    self.profile = 'DEFAULT'
            else:
                self.profile = self.goatshell_instance.get_profile(self.goatshell_instance.prefix)
            getLayout()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

        @self.add(Keys.F9)
        def handle_f9(event):
            if self.goatshell_instance.prefix == 'az':
                self.goatshell_instance.switch_to_next_subscription()  # Call the method on the Goatshell instance
            else:
                self.profile = self.goatshell_instance.get_profile(self.goatshell_instance.prefix)
            getLayout()
            self.app.invalidate()
            event.app.exit(result='re-prompt')

        @self.add(Keys.F10)
        def handle_f10(event):
            self.goatshell_instance.toggle_vi_mode()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

        @self.add(Keys.F12)
        def handle_f12(event):
            self.goatshell_instance.toggle_safety_mode()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.
